==> v10.6.0.zip: brewman/brewman/brewman/routers/recipe.py <==
import uuid

from datetime import date
from decimal import Decimal
from typing import List, Optional

import brewman.schemas.recipe as schemas
import brewman.schemas.recipe_item

from fastapi import APIRouter, Depends, HTTPException, Request, Security, status
from sqlalchemy import delete, desc, func, select
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Session, contains_eager

from ..core.security import get_current_active_user as get_user
from ..db.session import SessionFuture
from ..models.batch import Batch
from ..models.inventory import Inventory
from ..models.period import Period
from ..models.product import Product
from ..models.recipe import Recipe
from ..models.recipe_item import RecipeItem
from ..models.stock_keeping_unit import StockKeepingUnit
from ..models.voucher import Voucher
from ..schemas.product_sku import ProductSku
from ..schemas.user import UserToken
from .reports import report_finish_date, report_start_date


router = APIRouter()


@router.post("", response_model=schemas.Recipe)
def save(
    data: schemas.RecipeIn,
    request: Request,
    user: UserToken = Security(get_user, scopes=["recipes"]),
) -> schemas.Recipe:
    try:
        with SessionFuture() as db:
            recipe_sku: StockKeepingUnit = db.execute(
                select(StockKeepingUnit).where(StockKeepingUnit.id == data.sku.id_)
            ).scalar_one()

            if recipe_sku.product_id in set(i.product.id_ for i in data.items):
                raise HTTPException(
                    status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail="Product cannot also be an ingredient",
                )
            if len(data.items) == 0:
                raise HTTPException(
                    status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail="Recipe has no ingredients",
                )
            period: Period = db.execute(select(Period).where(Period.id == data.period.id_)).scalar_one()
            recipe = Recipe(
                sku=recipe_sku,
                recipe_yield=round(data.recipe_yield, 2),
                sale_price=round(data.sale_price, 2),
                period_id=data.period.id_,
                cost_price=Decimal(0),
            )
            recipe_cost = Decimal(0)
            for item in data.items:
                product: Product = db.execute(select(Product).where(Product.id == item.product.id_)).scalar_one()
                quantity = round(item.quantity, 2)
                if product.is_purchased:
                    ingredient_cost = get_purchased_product_cost(product, period.valid_from, period.valid_till, db)
                else:
                    ingredient_cost = get_sub_product_cost(product, period.id, db)
                    # ingredient_cost = get_purchased_product_cost(product.id, period.valid_from, period.valid_till, db)
                recipe_cost += round(ingredient_cost * quantity, 2)
                recipe.items.append(RecipeItem(product.id, quantity, ingredient_cost))

            recipe.cost_price = round(recipe_cost / recipe.recipe_yield, 2)
            recipe_sku.cost_price = round(recipe.cost_price / recipe.recipe_yield, 2)
            if recipe_sku.product.is_sold:
                recipe_sku.sale_price = recipe.sale_price

            db.add(recipe)
            for r_item in recipe.items:
                r_item.recipe_id = recipe.id
                db.add(r_item)

            check_recursion(recipe_sku.product_id, recipe, db)
            db.commit()
            return recipe_info(recipe)
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


@router.put("/{id_}", response_model=schemas.Recipe)
async def update_route(
    id_: uuid.UUID,
    data: schemas.RecipeIn,
    request: Request,
    user: UserToken = Security(get_user, scopes=["recipes"]),
) -> schemas.Recipe:
    try:
        with SessionFuture() as db:
            recipe: Recipe = db.execute(select(Recipe).where(Recipe.id == id_)).scalar_one()

            sku: StockKeepingUnit = db.execute(
                select(StockKeepingUnit).where(StockKeepingUnit.id == data.sku.id_)
            ).scalar_one()
            if sku.product_id in set(i.product.id_ for i in data.items):
                raise HTTPException(
                    status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail="Product cannot also be an ingredient",
                )

            period: Period = db.execute(select(Period).where(Period.id == data.period.id_)).scalar_one()
            recipe.sku = sku
            recipe.recipe_yield = round(data.recipe_yield, 2)
            recipe.sale_price = round(data.sale_price, 2)
            recipe.period = period

            recipe_cost = Decimal(0)

            for i in range(len(recipe.items), 0, -1):
                item = recipe.items[i - 1]
                index = next((idx for (idx, d) in enumerate(data.items) if d.id_ == item.id), None)
                product: Product
                if index is not None:
                    new_item = data.items.pop(index)
                    product = db.execute(select(Product).where(Product.id == new_item.product.id_)).scalar_one()
                    item.product_id = new_item.product.id_
                    item.quantity = round(new_item.quantity, 2)
                    if product.is_purchased:
                        ingredient_cost = get_purchased_product_cost(product, period.valid_from, period.valid_till, db)
                    else:
                        ingredient_cost = get_sub_product_cost(product, period.id, db)
                        # ingredient_cost = get_purchased_product_cost(product.id, period.valid_from, period.valid_till, db)
                    item.price = ingredient_cost
                    recipe_cost += round(ingredient_cost * item.quantity, 2)
                else:
                    recipe.items.remove(item)

            for d_item in data.items:
                product = db.execute(select(Product).where(Product.id == d_item.product.id_)).scalar_one()
                quantity = round(d_item.quantity, 2)
                if product.is_purchased:
                    ingredient_cost = get_purchased_product_cost(product, period.valid_from, period.valid_till, db)
                else:
                    ingredient_cost = get_sub_product_cost(product, period.id, db)
                    # ingredient_cost = get_purchased_product_cost(product.id, period.valid_from, period.valid_till, db)

                recipe_cost += round(ingredient_cost * quantity, 2)
                recipe.items.append(RecipeItem(product.id, quantity, ingredient_cost))

            recipe.cost_price = round(recipe_cost / recipe.recipe_yield, 2)
            sku.cost_price = round(recipe.cost_price / recipe.recipe_yield, 2)
            check_recursion(sku.product_id, recipe, db)
            db.commit()
            return recipe_info(recipe)
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


def check_recursion(product_id: uuid.UUID, recipe: Recipe, db: Session) -> None:
    for item in recipe.items:
        item_recipes: List[Recipe] = (
            db.execute(
                select(Recipe)
                .join(Recipe.sku)
                .join(Recipe.items)
                .where(
                    StockKeepingUnit.product_id == item.product_id,
                    Recipe.period_id == recipe.period_id,
                )
                .options(contains_eager(Recipe.sku), contains_eager(Recipe.items))
            )
            .unique()
            .scalars()
            .all()
        )
        for sub_recipe in item_recipes:
            if sub_recipe.sku.product_id == product_id:
                raise HTTPException(
                    status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail="Recipe recursion. Some ingredient recipe contains parent recipe.",
                )
            check_recursion(product_id, sub_recipe, db)


def get_purchased_product_cost(product: Product, start_date: date, finish_date: date, db: Session) -> Decimal:
    query = (
        select(Batch)
        .join(Batch.inventories)
        .join(Inventory.voucher)
        .join(Batch.sku)
        .join(StockKeepingUnit.product)
        .where(Product.id == product.id, Voucher.date >= start_date, Voucher.date <= finish_date)
    )
    batches: List[Batch] = db.execute(query).unique().scalars().all()
    skus: List[StockKeepingUnit] = (
        db.execute(select(StockKeepingUnit).where(StockKeepingUnit.product_id == product.id)).scalars().all()
    )
    price: List[Decimal] = []
    for batch in batches:
        price.append(batch.rate / batch.sku.fraction / batch.sku.product_yield)
    else:
        for sku in skus:
            price.append(sku.cost_price / sku.fraction / sku.product_yield)
    return round(sum(price) / len(price), 5)  # type: ignore[return-value]


def get_sub_product_cost(product: Product, period_id: uuid.UUID, db: Session) -> Decimal:
    recipe: Optional[Recipe] = (
        db.execute(
            select(Recipe)
            .join(Recipe.sku)
            .where(
                StockKeepingUnit.product_id == product.id,
                Recipe.period_id == period_id,
            )
        )
        .scalars()
        .one_or_none()
    )

    if recipe is None:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Ingredient {product.name} is not bought and also does not have a recipe for the period.",
        )

    return round(recipe.cost_price / recipe.sku.fraction / recipe.sku.product_yield, 5)  # type: ignore[no-any-return]


@router.delete("/{id_}", response_model=None)
def delete_route(
    id_: uuid.UUID,
    request: Request,
    user: UserToken = Security(get_user, scopes=["recipes"]),
) -> None:
    with SessionFuture() as db:
        recipe: Recipe = db.execute(select(Recipe).where(Recipe.id == id_)).scalar_one()
        recipe_ids: List[uuid.UUID] = (
            db.execute(
                select(func.distinct(RecipeItem.recipe_id)).where(
                    RecipeItem.product_id
                    == select(StockKeepingUnit.product_id)
                    .where(StockKeepingUnit.id == select(Recipe.sku_id).where(Recipe.id == id_).scalar_subquery())
                    .scalar_subquery(),
                    RecipeItem.recipe_id.in_(select(Recipe.id).where(Recipe.period_id == recipe.period_id)),
                )
            )
            .scalars()
            .all()
        )
        if len(recipe_ids) > 0:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="This recipe has other recipes dependent on it and cannot be deleted",
            )

        db.execute(delete(RecipeItem).where(RecipeItem.recipe_id == id_))
        db.execute(delete(Recipe).where(Recipe.id == id_))

        db.commit()


@router.get("", response_model=schemas.RecipeBlank)
def show_blank(
    request: Request,
    user: UserToken = Security(get_user, scopes=["recipes"]),
) -> schemas.RecipeBlank:
    with SessionFuture() as db:
        period_id: uuid.UUID = db.execute(select(Period.id).order_by(desc(Period.valid_till)).limit(1)).scalar_one()
    return recipe_blank(period_id)


@router.get("/list", response_model=List[schemas.Recipe])
async def show_list(
    p: uuid.UUID,  # period_id
    user: UserToken = Depends(get_user),
) -> List[schemas.Recipe]:
    with SessionFuture() as db:
        list_: List[Recipe] = (
            db.execute(select(Recipe).join(Recipe.sku).where(Recipe.period_id == p).order_by(Recipe.sku_id))
            .scalars()
            .all()
        )
        return [
            schemas.Recipe(
                id=item.id,
                period=schemas.PeriodLink(id=item.period.id),
                sku=schemas.ProductLink(id=item.sku.id, name=f"{item.sku.product.name} ({item.sku.units})"),
                recipeYield=item.recipe_yield,
                costPrice=item.cost_price,
                salePrice=item.sale_price,
                notes=str(item.sku.product.product_group_id),
                items=[],
            )
            for item in list_
        ]


@router.get("/ingredient-details/{id_}", response_model=ProductSku)
def show_ingredient_id(
    id_: uuid.UUID,
    start_date: Optional[date] = Depends(report_start_date),
    finish_date: Optional[date] = Depends(report_finish_date),
    user: UserToken = Security(get_user, scopes=["recipes"]),
) -> ProductSku:
    with SessionFuture() as db:
        query = (
            select(Batch)
            .join(Batch.inventories)
            .join(Inventory.voucher)
            .join(Batch.sku)
            .join(StockKeepingUnit.product)
            .where(Product.id == id_)
        )
        if start_date is not None:
            query = query.where(Voucher.date >= start_date)
        if finish_date is not None:
            query = query.where(Voucher.date <= finish_date)
        batches: List[Batch] = db.execute(query).unique().scalars().all()
        product: Product = db.execute(select(Product).join(Product.skus).where(Product.id == id_)).unique().scalar_one()
        price: List[Decimal] = []
        for batch in batches:
            price.append(batch.rate / batch.sku.fraction / batch.sku.product_yield)
        else:
            for sku in product.skus:
                price.append(sku.cost_price / sku.fraction / sku.product_yield)
        return ProductSku(
            id=id_,
            name=product.name,
            fractionUnits=product.fraction_units,
            costPrice=round(sum(price) / len(price), 2),
            salePrice=0,
            isRateContracted=False,
        )


@router.get("/{id_}", response_model=schemas.Recipe)
def show_id(
    id_: uuid.UUID,
    user: UserToken = Security(get_user, scopes=["recipes"]),
) -> schemas.Recipe:
    with SessionFuture() as db:
        item: Recipe = db.execute(select(Recipe).where(Recipe.id == id_)).scalar_one()
        return recipe_info(item)


def recipe_info(recipe: Recipe) -> schemas.Recipe:
    return schemas.Recipe(
        id=recipe.id,
        sku=schemas.ProductLink(
            id=recipe.sku_id,
            name=f"{recipe.sku.product.name} ({recipe.sku.units})",
        ),
        period=schemas.PeriodLink(id=recipe.period_id),
        recipeYield=recipe.recipe_yield,
        salePrice=recipe.sale_price,
        costPrice=recipe.cost_price,
        notes="",
        items=[
            brewman.schemas.recipe_item.RecipeItem(
                id=item.id,
                product=schemas.ProductLink(
                    id=item.product.id,
                    name=f"{item.product.name} ({item.product.fraction_units})",
                ),
                quantity=round(item.quantity, 2),
                price=round(item.price, 5),
            )
            for item in recipe.items
        ],
    )


def recipe_blank(period_id: uuid.UUID) -> schemas.RecipeBlank:
    return schemas.RecipeBlank(
        period=schemas.PeriodLink(id=period_id),
        recipeYield=1,
        costPrice=0,
        salePrice=0,
        isLocked=False,
        notes="",
        items=[],
    )

==> v11.0.1.zip: brewman/brewman/brewman/routers/recipe.py <==
import uuid

from datetime import date
from decimal import Decimal
from typing import List, Optional

import brewman.schemas.recipe as schemas
import brewman.schemas.recipe_item

from fastapi import APIRouter, Depends, HTTPException, Request, Security, status
from sqlalchemy import delete, desc, func, select
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Session, contains_eager

from ..core.security import get_current_active_user as get_user
from ..db.session import SessionFuture
from ..models.batch import Batch
from ..models.inventory import Inventory
from ..models.period import Period
from ..models.product import Product
from ..models.recipe import Recipe
from ..models.recipe_item import RecipeItem
from ..models.stock_keeping_unit import StockKeepingUnit
from ..models.voucher import Voucher
from ..schemas.product_sku import ProductSku
from ..schemas.user import UserToken
from .reports import report_finish_date, report_start_date


router = APIRouter()


@router.post("", response_model=schemas.Recipe)
def save(
    data: schemas.RecipeIn,
    request: Request,
    user: UserToken = Security(get_user, scopes=["recipes"]),
) -> schemas.Recipe:
    try:
        with SessionFuture() as db:
            recipe_sku: StockKeepingUnit = db.execute(
                select(StockKeepingUnit).where(StockKeepingUnit.id == data.sku.id_)
            ).scalar_one()

            if recipe_sku.product_id in set(i.product.id_ for i in data.items):
                raise HTTPException(
                    status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail="Product cannot also be an ingredient",
                )
            if len(data.items) == 0:
                raise HTTPException(
                    status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail="Recipe has no ingredients",
                )
            period: Period = db.execute(select(Period).where(Period.id == data.period.id_)).scalar_one()
            recipe = Recipe(
                sku=recipe_sku,
                recipe_yield=round(data.recipe_yield, 2),
                sale_price=round(data.sale_price, 2),
                period_id=data.period.id_,
                cost_price=Decimal(0),
            )
            recipe_cost = Decimal(0)
            for item in data.items:
                product: Product = db.execute(select(Product).where(Product.id == item.product.id_)).scalar_one()
                quantity = round(item.quantity, 2)
                if product.is_purchased:
                    ingredient_cost = get_purchased_product_cost(product, period.valid_from, period.valid_till, db)
                else:
                    ingredient_cost = get_sub_product_cost(product, period.id, db)
                    # ingredient_cost = get_purchased_product_cost(product.id, period.valid_from, period.valid_till, db)
                recipe_cost += round(ingredient_cost * quantity, 2)
                recipe.items.append(RecipeItem(product.id, quantity, ingredient_cost))

            recipe.cost_price = round(recipe_cost / recipe.recipe_yield, 2)
            recipe_sku.cost_price = round(recipe.cost_price / recipe.recipe_yield, 2)
            if recipe_sku.product.is_sold:
                recipe_sku.sale_price = recipe.sale_price

            db.add(recipe)
            for r_item in recipe.items:
                r_item.recipe_id = recipe.id
                db.add(r_item)

            check_recursion(recipe_sku.product_id, recipe, db)
            db.commit()
            return recipe_info(recipe)
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


@router.put("/{id_}", response_model=schemas.Recipe)
async def update_route(
    id_: uuid.UUID,
    data: schemas.RecipeIn,
    request: Request,
    user: UserToken = Security(get_user, scopes=["recipes"]),
) -> schemas.Recipe:
    try:
        with SessionFuture() as db:
            recipe: Recipe = db.execute(select(Recipe).where(Recipe.id == id_)).scalar_one()

            sku: StockKeepingUnit = db.execute(
                select(StockKeepingUnit).where(StockKeepingUnit.id == data.sku.id_)
            ).scalar_one()
            if sku.product_id in set(i.product.id_ for i in data.items):
                raise HTTPException(
                    status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail="Product cannot also be an ingredient",
                )

            period: Period = db.execute(select(Period).where(Period.id == data.period.id_)).scalar_one()
            recipe.sku = sku
            recipe.recipe_yield = round(data.recipe_yield, 2)
            recipe.sale_price = round(data.sale_price, 2)
            recipe.period = period

            recipe_cost = Decimal(0)

            for i in range(len(recipe.items), 0, -1):
                item = recipe.items[i - 1]
                index = next((idx for (idx, d) in enumerate(data.items) if d.id_ == item.id), None)
                product: Product
                if index is not None:
                    new_item = data.items.pop(index)
                    product = db.execute(select(Product).where(Product.id == new_item.product.id_)).scalar_one()
                    item.product_id = new_item.product.id_
                    item.quantity = round(new_item.quantity, 2)
                    if product.is_purchased:
                        ingredient_cost = get_purchased_product_cost(product, period.valid_from, period.valid_till, db)
                    else:
                        ingredient_cost = get_sub_product_cost(product, period.id, db)
                        # ingredient_cost = get_purchased_product_cost(product.id, period.valid_from, period.valid_till, db)
                    item.price = ingredient_cost
                    recipe_cost += round(ingredient_cost * item.quantity, 2)
                else:
                    recipe.items.remove(item)

            for d_item in data.items:
                product = db.execute(select(Product).where(Product.id == d_item.product.id_)).scalar_one()
                quantity = round(d_item.quantity, 2)
                if product.is_purchased:
                    ingredient_cost = get_purchased_product_cost(product, period.valid_from, period.valid_till, db)
                else:
                    ingredient_cost = get_sub_product_cost(product, period.id, db)
                    # ingredient_cost = get_purchased_product_cost(product.id, period.valid_from, period.valid_till, db)

                recipe_cost += round(ingredient_cost * quantity, 2)
                recipe.items.append(RecipeItem(product.id, quantity, ingredient_cost))

            recipe.cost_price = round(recipe_cost / recipe.recipe_yield, 2)
            sku.cost_price = round(recipe.cost_price / recipe.recipe_yield, 2)
            check_recursion(sku.product_id, recipe, db)
            db.commit()
            return recipe_info(recipe)
    except SQLAlchemyError as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(e),
        )


def check_recursion(product_id: uuid.UUID, recipe: Recipe, db: Session) -> None:
    for item in recipe.items:
        item_recipes: List[Recipe] = (
            db.execute(
                select(Recipe)
                .join(Recipe.sku)
                .join(Recipe.items)
                .where(
                    StockKeepingUnit.product_id == item.product_id,
                    Recipe.period_id == recipe.period_id,
                )
                .options(contains_eager(Recipe.sku), contains_eager(Recipe.items))
            )
            .unique()
            .scalars()
            .all()
        )
        for sub_recipe in item_recipes:
            if sub_recipe.sku.product_id == product_id:
                raise HTTPException(
                    status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail="Recipe recursion. Some ingredient recipe contains parent recipe.",
                )
            check_recursion(product_id, sub_recipe, db)


def get_purchased_product_cost(product: Product, start_date: date, finish_date: date, db: Session) -> Decimal:
    query = (
        select(Batch)
        .join(Batch.inventories)
        .join(Inventory.voucher)
        .join(Batch.sku)
        .join(StockKeepingUnit.product)
        .where(Product.id == product.id, Voucher.date >= start_date, Voucher.date <= finish_date)
    )
    batches: List[Batch] = db.execute(query).unique().scalars().all()
    skus: List[StockKeepingUnit] = (
        db.execute(select(StockKeepingUnit).where(StockKeepingUnit.product_id == product.id)).scalars().all()
    )
    price: List[Decimal] = []
    for batch in batches:
        price.append(batch.rate / batch.sku.fraction / batch.sku.product_yield)
    else:
        for sku in skus:
            price.append(sku.cost_price / sku.fraction / sku.product_yield)
    return round(sum(price) / len(price), 5)  # type: ignore[return-value]


def get_sub_product_cost(product: Product, period_id: uuid.UUID, db: Session) -> Decimal:
    recipe: Optional[Recipe] = (
        db.execute(
            select(Recipe)
            .join(Recipe.sku)
            .where(
                StockKeepingUnit.product_id == product.id,
                Recipe.period_id == period_id,
            )
        )
        .scalars()
        .one_or_none()
    )

    if recipe is None:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Ingredient {product.name} is not bought and also does not have a recipe for the period.",
        )

    return round(recipe.cost_price / recipe.sku.fraction / recipe.sku.product_yield, 5)


@router.delete("/{id_}", response_model=None)
def delete_route(
    id_: uuid.UUID,
    request: Request,
    user: UserToken = Security(get_user, scopes=["recipes"]),
) -> None:
    with SessionFuture() as db:
        recipe: Recipe = db.execute(select(Recipe).where(Recipe.id == id_)).scalar_one()
        recipe_ids: List[uuid.UUID] = (
            db.execute(
                select(func.distinct(RecipeItem.recipe_id)).where(
                    RecipeItem.product_id
                    == select(StockKeepingUnit.product_id)
                    .where(StockKeepingUnit.id == select(Recipe.sku_id).where(Recipe.id == id_).scalar_subquery())
                    .scalar_subquery(),
                    RecipeItem.recipe_id.in_(select(Recipe.id).where(Recipe.period_id == recipe.period_id)),
                )
            )
            .scalars()
            .all()
        )
        if len(recipe_ids) > 0:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="This recipe has other recipes dependent on it and cannot be deleted",
            )

        db.execute(delete(RecipeItem).where(RecipeItem.recipe_id == id_))
        db.execute(delete(Recipe).where(Recipe.id == id_))

        db.commit()


@router.get("", response_model=schemas.RecipeBlank)
def show_blank(
    request: Request,
    user: UserToken = Security(get_user, scopes=["recipes"]),
) -> schemas.RecipeBlank:
    with SessionFuture() as db:
        period_id: uuid.UUID = db.execute(select(Period.id).order_by(desc(Period.valid_till)).limit(1)).scalar_one()
    return recipe_blank(period_id)


@router.get("/list", response_model=List[schemas.Recipe])
async def show_list(
    p: uuid.UUID,  # period_id
    user: UserToken = Depends(get_user),
) -> List[schemas.Recipe]:
    with SessionFuture() as db:
        list_: List[Recipe] = (
            db.execute(select(Recipe).join(Recipe.sku).where(Recipe.period_id == p).order_by(Recipe.sku_id))
            .scalars()
            .all()
        )
        return [
            schemas.Recipe(
                id=item.id,
                period=schemas.PeriodLink(id=item.period.id),
                sku=schemas.ProductLink(id=item.sku.id, name=f"{item.sku.product.name} ({item.sku.units})"),
                recipeYield=item.recipe_yield,
                costPrice=item.cost_price,
                salePrice=item.sale_price,
                notes=str(item.sku.product.product_group_id),
                items=[],
            )
            for item in list_
        ]


@router.get("/ingredient-details/{id_}", response_model=ProductSku)
def show_ingredient_id(
    id_: uuid.UUID,
    start_date: Optional[date] = Depends(report_start_date),
    finish_date: Optional[date] = Depends(report_finish_date),
    user: UserToken = Security(get_user, scopes=["recipes"]),
) -> ProductSku:
    with SessionFuture() as db:
        query = (
            select(Batch)
            .join(Batch.inventories)
            .join(Inventory.voucher)
            .join(Batch.sku)
            .join(StockKeepingUnit.product)
            .where(Product.id == id_)
        )
        if start_date is not None:
            query = query.where(Voucher.date >= start_date)
        if finish_date is not None:
            query = query.where(Voucher.date <= finish_date)
        batches: List[Batch] = db.execute(query).unique().scalars().all()
        product: Product = db.execute(select(Product).join(Product.skus).where(Product.id == id_)).unique().scalar_one()
        price: List[Decimal] = []
        for batch in batches:
            price.append(batch.rate / batch.sku.fraction / batch.sku.product_yield)
        else:
            for sku in product.skus:
                price.append(sku.cost_price / sku.fraction / sku.product_yield)
        return ProductSku(
            id=id_,
            name=product.name,
            fractionUnits=product.fraction_units,
            costPrice=round(sum(price) / len(price), 2),
            salePrice=0,
            isRateContracted=False,
        )


@router.get("/{id_}", response_model=schemas.Recipe)
def show_id(
    id_: uuid.UUID,
    user: UserToken = Security(get_user, scopes=["recipes"]),
) -> schemas.Recipe:
    with SessionFuture() as db:
        item: Recipe = db.execute(select(Recipe).where(Recipe.id == id_)).scalar_one()
        return recipe_info(item)


def recipe_info(recipe: Recipe) -> schemas.Recipe:
    return schemas.Recipe(
        id=recipe.id,
        sku=schemas.ProductLink(
            id=recipe.sku_id,
            name=f"{recipe.sku.product.name} ({recipe.sku.units})",
        ),
        period=schemas.PeriodLink(id=recipe.period_id),
        recipeYield=recipe.recipe_yield,
        salePrice=recipe.sale_price,
        costPrice=recipe.cost_price,
        notes="",
        items=[
            brewman.schemas.recipe_item.RecipeItem(
                id=item.id,
                product=schemas.ProductLink(
                    id=item.product.id,
                    name=f"{item.product.name} ({item.product.fraction_units})",
                ),
                quantity=round(item.quantity, 2),
                price=round(item.price, 5),
            )
            for item in recipe.items
        ],
    )


def recipe_blank(period_id: uuid.UUID) -> schemas.RecipeBlank:
    return schemas.RecipeBlank(
        period=schemas.PeriodLink(id=period_id),
        recipeYield=1,
        costPrice=0,
        salePrice=0,
        isLocked=False,
        notes="",
        items=[],
    )
